"""Calcium"""
from . import client
import warnings
from . import utils
from PIL import Image
from typing import List
import io

receive_fname = ''
receive_count = 0
receive_data = []

CHUNK_SIZE = 1000000

counter = 0
class FOV:
    def __init__(self, position=[0.0, 0.0, 0.0], offset=0.0, texture_file=None):
        self.create()

        position = utils.sanitize_vector3(position)
        self.set_position(position)
        self.set_offset(offset)
        if texture_file: self.set_texture(texture_file)

    def create(self):
        """Creates FOVs

        Parameters
        ---------- 
        none

        Examples
        >>>f1 = urchin.fov.FOV()
        """
        global counter
        counter += 1
        self.id = 'fov' + str(counter)
        client.sio.emit('CreateFOV',[self.id])
        self.in_unity = True

    def delete(self):
        """Deletes fovs

        Parameters
        ---------- 
        references object being deleted

        Examples
        >>>f1.delete()
        """
        client.sio.emit('DeleteFOV',[self.id])
        self.in_unity = False

    def set_position(self,position):
        """Set the position of fov in ml/ap/dv coordinates relative to the CCF (0,0,0) point

        Parameters
        ---------- 
        position : list of three floats
        	vertex positions of the fov relative to the CCF point

        Examples
        --------
        >>>f1.set_position([2,2,2])
        """
        if self.in_unity == False:
            raise Exception("fov does not exist in Unity, call create method first.")

        position = utils.sanitize_vector3(position)
        self.position = position
        client.sio.emit('SetFOVPos',{self.id: position})
    
    def set_texture(self,texture_file):
        """Set the texture of fov from an image

        Parameters
        ----------
        texture_file : str, file name of FOV image

        Examples
        --------
        >>>f1.set_texture('fov1.png')
        """
        if self.in_unity == False:
            raise Exception("fov does not exist in Unity, call create method first.")

        texture_file = utils.sanitize_string(texture_file)
        self.texture_file=texture_file

        img = Image.open(texture_file)
        # Convert img to bytes
        img_bytes = io.BytesIO()
        img.save(img_bytes,format=img.format)
        img_bytes.seek(0)
        # Split bytes into chunks
        chunks = []
        chunk = []
        while chunk:
            chunk = img_bytes.read(CHUNK_SIZE)
            chunks.append(chunk)
        # Send img by chunk
        for i,chunk in enumerate(chunks):
            immediate_apply = True if i==len(chunks)-1 else False
            client.sio.emit('SetFOVTextureDataMeta', [self.id,i,immediate_apply])
            client.sio.emit('SetFOVTextureData',chunk)

    def set_offset(self,fov_offsets):
        client.sio.emit('SetFOVOffset',fov_offsets)


def create(num_fovs):
    """Create fov objects

    Note: fovs must be created before setting other values

    Parameters
    ----------
    num_fovs : int
        number of new fov objects

    Examples
    --------
    >>> fovs = urchin.fovs.create(3)
    """
    fov_ids = []
    for i in range(num_fovs):
        fov = FOV()
        fov_ids.append(fov.id)
    return fov_ids

def delete(fovs_list):
    """Delete fov objects

    Parameters
    ----------
    fov_names : list of fov objects
    list of fovs being deleted

    Examples
    --------
    >>> urchin.fovs.delete()
    """
    fovs_list = utils.sanitize_list(fovs_list)
    for fov in fovs_list:
        if fov.in_unity:
            fov.delete()
        else:
            warnings.warn(f"fov with id {fov.id} does not exist in Unity, call create method first.")

    fovs_ids = [x.id for x in fovs_list]
    client.sio.emit('DeleteFOVs', fovs_ids)

def set_positions(fovs_list, positions_list):
    """Set the position of fov in ml/ap/dv coordinates relative to the CCF (0,0,0) point

    Parameters
    ----------
    fovs_list : list of fov objects
        list of fovs being moved
    positions : list of list of three floats
        list of positions of fovs

    Examples
    --------
    >>> urchin.fovs.set_positions([f1,f2,f3], [[1,1,1],[2,2,2],[3,3,3]])
    """
    fovs_list = utils.sanitize_list(fovs_list)
    positions_list = utils.sanitize_list(positions_list)

    for fov,position in zip(fovs_list, positions_list):
        if fov.in_unity:
            fov.set_position(position)
        else:
            warnings.warn(f"fov with id {fov.id} does not exist in Unity, call create method first.")

def set_textures(fovs_list, texture_files_list):
    """Set the position of fov in ml/ap/dv coordinates relative to the CCF (0,0,0) point

    Parameters
    ----------
    fovs_list : list of fov objects
        list of fovs being moved
    positions : list of list of three floats
        list of positions of fovs

    Examples
    --------
    >>> urchin.fovs.set_textures([f1,f2,f3], ['fov1.png','fov'])
    """
    fovs_list = utils.sanitize_list(fovs_list)
    texture_files_list = utils.sanitize_list(texture_files_list)

    for fov,texture in zip(fovs_list,texture_files_list):
        if fov.in_unity:
            fov.set_texture(texture)
        else:
            warnings.warn(f"fov with id {fov.id} does not exist in Unity, call create method first.")


import webbrowser
import os
from IPython.display import Javascript

from . import client

def is_running_in_colab():
	return True #'google.colab' in str(get_ipython())

def setup(verbose = True, localhost = False, standalone = False):
	"""Connect the Unity Renderer for Neuroscience Python API to the web-based (or standalone) viewer

	Parameters
	----------
	localhost : bool, optional
		connect to a local development server rather than the remote server, by default False
	standalone : bool, optional
		connect to a standalone Desktop build rather than the web-based Brain Viewer, by default False
	"""

	if client.connected():
		print(f'(urchin) Client is already connected. Use ID: {client.ID}')
		return
		
	log_messages = verbose

	if localhost:
		client.sio.connect('http://localhost:5000')
	else:
		client.sio.connect('https://urchin-commserver.herokuapp.com/')

	if not standalone:
		#To open browser window:
		url = f'https://data.virtualbrainlab.org/Urchin/?ID={client.ID}'
		if not is_running_in_colab():
			webbrowser.open(url)
		else:
			# Specify window features
			window_features = "width=1200,height=800,toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,resizable=yes"
			# Use the window.open function with window features
			javascript_code = f'window.open("{url}", "_blank", "{window_features}");'
			# Display the JavaScript code to open the new window
			display(Javascript(javascript_code))

			

######################
# CLEAR #
######################

def clear():
	"""Clear the renderer scene of all objects
	"""
	client.sio.emit('Clear', 'all')

def clear_neurons():
	"""Clear all neuron objects
	"""
	client.sio.emit('Clear', 'neurons')

def clear_probes():
	"""Clear all probe objects
	"""
	client.sio.emit('Clear', 'probes')

def clear_areas():
	"""Clear all CCF area models
	"""
	client.sio.emit('Clear', 'areas')

def clear_volumes():
	"""Clear all 3D volumes
	"""
	client.sio.emit('Clear', 'volumes')

def clear_texts():
	"""Clear all text
	"""
	client.sio.emit('Clear', 'texts')

def clear_primitives():
	"""Clear all primitives
	"""
	client.sio.emit('Clear','primitives')


"""Text"""

from . import client
import warnings
from . import utils

## Text renderer

counter = 0
class Text:
  def __init__(self, text = "", color = "#FFFFFF", size = 12, positon = [0,0]):
    self.create()

    text = utils.sanitize_string(text)
    self.text = text
    client.sio.emit('SetTextText',{self.id: text})

    color = utils.sanitize_color(color)
    self.color = color
    client.sio.emit('SetTextColors',{self.id: color})

    position = utils.sanitize_list(position)
    self.position = position
    client.sio.emit('SetTextPositions',{self.id: position})


  def create(self):
    """Create a text object

    Parameters
    ----------
    none
        
    Examples
    --------
    >>> t1 = urchin.text.Text()
    """
    global counter
    counter +=1
    self.id = 't' + str(counter)
    client.sio.emit('CreateText',[self.id])
    self.in_unity = True
  
  def delete(self):
    """Delete a text object

    Parameters
    ----------
    references object being deleted
        
    Examples
    --------
    >>> t1.delete()
    """
    client.sio.emit('DeleteText',[self.id])
    self.in_unity = False

  def set_text(self, text_text):
    """Set the text in a set of text objects

    Parameters
    ----------
    text : string
      text to be displayed

    Examples
    --------
    >>> t1.set_text('test text')
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    text_text = utils.sanitize_string(text_text)
    self.text = text_text
    client.sio.emit('SetTextText',{self.id: text_text})

  def set_color(self,text_color):
    """Set the color of a set of text objects

    Parameters
    ----------
    text_colors : string hex color
        hex colors as strings
        
    Examples
    --------
    >>> t1.set_color('#FF0000')
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    text_color = utils.sanitize_color(text_color)
    self.color = text_color
    client.sio.emit('SetTextColors',{self.id: text_color})

  def set_size(self,text_size):
    """Set the font size of a set of text objects

    Parameters
    ----------
    text_sizes : int
        font sizes
        
    Examples
    --------
    >>> t1.set_size(12)
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    self.size = text_size
    client.sio.emit('SetTextSizes',{self.id: text_size})

  def set_position(self,text_pos):
    """Set the positions of a set of text objects in UI canvas space
    Bottom left corner is [-1,-1], top right [1,1]

    Text is anchored at the top left corner of its text box.

    Parameters
    ----------
    text_pos : list of two floats
        canvas positions relative to the center
        
    Examples
    --------
    >>> t1.set_position([400, 300])
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    position = utils.sanitize_list(position)
    self.position = position
    client.sio.emit('SetTextPositions',{self.id: position})



"""Camera"""

from . import client
from . import utils

from PIL import Image
import io
		  
receive_fname = ''
receive_count = 0
receive_data = []

def receive_camera_img_meta(data):
	global receive_count, receive_data
	receive_count = int(data)
	print('(Camera receive meta) ' + str(data))

def receive_camera_img(data):
  global receive_count, receive_data
	
  print(f'(Camera) received {str(len(data))} bytes')
  receive_data.append(data)
  receive_count -= 1

  if (receive_count == 0):
    data_bytes = b''.join(receive_data)
    receive_data = []
    Image.open(io.BytesIO(data_bytes)).save(receive_fname)
    print('(Camera received all data)')

## Camera renderer
counter = 0

class Camera:
	def __init__(self, target = [0,0,0], position = [0,0,0], preserve_target = True, rotation = [0,0,0], zoom = 1, pan_x = 3, pan_y = 4, mode = "orthographic", controllable = False, main = False):
		self.create(main=main)
		#in theory, the target value can stand for either coordinate or area? (and taking coordinate as default)
		# target_coordinate = utils.sanitize_vector3(target)
		# self.target = target_coordinate
		# client.sio.emit('SetCameraTarget', {self.id: self.target})

		# position = utils.sanitize_vector3(position)
		# self.position = position
		# packet = position.copy()
		# packet.append(preserve_target)
		# client.sio.emit('SetCameraPosition', {self.id: packet})

		# rotation = utils.sanitize_vector3(rotation)
		# self.rotation = rotation
		# client.sio.emit('SetCameraRotation', {self.id: self.rotation})

		# self.zoom = zoom
		# client.sio.emit('SetCameraZoom', {self.id: self.zoom})

		# # target_area = utils.sanitize_string(target_area)
		# # self.target_area = target_area
		# # client.sio.emit('SetCameraTargetArea', self.target_area)

		# self.pan = [pan_x, pan_y]
		# client.sio.emit('SetCameraPan', {self.id: self.pan})

		# self.mode = mode
		# client.sio.emit('SetCameraMode', {self.id: self.mode})

	def create(self,main):
		"""Creates camera
		
		Parameters
		---------- 
		none

		Examples
		>>>c1 = urchin.camera.Camera()
		"""
		if(main):
			self.id = "main"
			self.in_unity = True
		else:
			global counter
			counter += 1
			self.id = str(counter)
			client.sio.emit('CreateCamera', [self.id])
			self.in_unity = True

	def delete(self):
		"""Deletes camera
		
		Parameters
		---------- 
		references object being deleted

		Examples
		>>>c1.delete()
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		client.sio.emit('DeleteCamera', [self.id])
		self.in_unity = False

	def set_target_coordinate(self,camera_target_coordinate):
		"""Set the camera target coordinate in CCF space in um relative to CCF (0,0,0), without moving the camera. Coordinates can be negative or larger than the CCF space (11400,13200,8000)

		Parameters
		----------
		camera_target_coordinate : float list
		  list of coordinates in ml/ap/dv in um

		Examples
		--------
		>>>c1.set_target_coordinate([500,1500,1000])
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		camera_target_coordinate = utils.sanitize_vector3(camera_target_coordinate)
		self.target = camera_target_coordinate
		client.sio.emit('SetCameraTarget', {self.id: camera_target_coordinate})

	def set_position(self, position, preserve_target = True):
		"""Set the camera position in CCF space in um relative to CCF (0,0,0), coordinates can be outside CCF space. 

		Parameters
		----------
		position : float list
			list of coordinates in ml/ap/dv in um
		preserve_target : bool, optional
			when True keeps the camera aimed at the current target, when False preserves the camera rotation, by default True
		
		Examples
		--------
		>>> c1.set_position([500,1500,1000])
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		position = utils.sanitize_vector3(position)
		self.position = position
		packet = position.copy()
		packet.append(preserve_target)
		client.sio.emit('SetCameraPosition', {self.id: packet})

	def set_rotation(self, rotation):
		"""Set the camera rotation (pitch, yaw, spin). The camera is locked to a target, so this rotation rotates around the target.

		Rotations are applied in order: pitch, then yaw, then spin.

		Parameters
		----------
		rotation : float list
			list of euler angles to set the camera rotation in (pitch, yaw, spin)

		Examples
		--------
		>>> c1.set_rotation([0,0,0])
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		rotation = utils.sanitize_vector3(rotation)
		self.rotation = rotation
		client.sio.emit('SetCameraRotation', {self.id: rotation})

	def set_rotation_axial(self, above=True):
		"""Set the camera to the standard axial view. Does not change the window, only the rotation.
		Parameters
		----------
		above : boolean
			default true, the above view

		Examples
		--------
		>>> c1.set_rotation_axial(True)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		if above:
			rotation = [0,0,0]
		else:
			rotation = [0,-180,0]
		rotation = utils.sanitize_vector3(rotation)
		self.rotation = rotation
		client.sio.emit('SetCameraRotation', {self.id: rotation})

	
	def set_rotation_coronal(self, back=True):
		"""Set the camera to the standard coronal view. Does not change the window, only the rotation.
		Parameters
		----------
		bool : boolean
			default true, the back view

		Examples
		--------
		>>> c1.set_rotation_coronal(True)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		if back:
			rotation = [90,0,0]
		else:
			rotation = [90,-180,0]
		rotation = utils.sanitize_vector3(rotation)
		self.rotation = rotation
		client.sio.emit('SetCameraRotation', {self.id: rotation})

	
	def set_rotation_sagittal(self, left=True):
		"""Set the camera to the standard sagittal view. Does not change the window, only the rotation.
		Parameters
		----------
		bool : boolean
			default true, the left view

		Examples
		--------
		>>> c1.set_rotation_sagittal(True)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		if left:
			rotation = [90,-90,0]
		else:
			rotation = [90,90,0]
		rotation = utils.sanitize_vector3(rotation)
		self.rotation = rotation
		client.sio.emit('SetCameraRotation', {self.id: rotation})



	def set_zoom(self,zoom):
		"""Set the camera zoom. 

		Parameters
		----------
		zoom : float	
			camera zoom parameter

		Examples
		--------
		>>> c1.set_zoom(1.0)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.zoom = zoom
		client.sio.emit('SetCameraZoom', {self.id: zoom})

	def set_target_area(self, camera_target_area):
		"""Set the camera rotation to look towards a target area

		Note: some long/curved areas have mesh centers that aren't the 'logical' center. For these areas, calculate a center yourself and use set_camera_target.

		Parameters
		----------
		camera_target_area : string
			area ID or acronym, append "-lh" or "-rh" for one-sided meshes
		Examples
		--------
		>>> c1.set_target_area("grey-l") 
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		camera_target_area
		self.target = camera_target_area
		client.sio.emit('SetCameraTargetArea', {self.id: camera_target_area})

	def set_pan(self,pan_x, pan_y):
		"""Set camera pan coordinates

		Parameters
		----------
		pan_x : float
			x coordinate
		pan_y : float
			y coordinate
		
		Examples
		--------
		>>> c1.set_pan(3.0, 4.0)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.pan = [pan_x, pan_y]
		client.sio.emit('SetCameraPan',{self.id: self.pan})

	def set_mode(self, mode):
		"""Set camera perspective mode

		Parameters
		----------
		mode : string
		mode options "perspective" or "orthographic" (default)
		
		Examples
		--------
		>>> c1.set_mode('perspective')
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		self.mode = mode
		client.sio.emit('SetCameraMode', {self.id: mode})

	def set_controllable(self):
		"""Sets camera to controllable

		Parameters
		----------
		self
		
		Examples
		--------
		>>> c1.set_controllable()
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		self.controllable = True
		client.sio.emit('SetCameraControl', self.id)
		


	# def capture_image(filename):
	# 	""" Capture a full screenshot and save to filename
		
	# 	Note: only supports PNG filetypes, for now
		
	# 	Examples
	# 	--------
	# 	>>> urn.capture_image('./image.png')
	# 	"""
	# 	global receive_fname
	# 	receive_fname = filename
	# 	client.sio.emit('RequestCameraImg')

def set_light_rotation(angles):
    angles = utils.sanitize_vector3(angles)
    print(angles)
    print(isinstance(angles,list))
    client.sio.emit('SetLightRotation', angles)

def set_light_camera(camera_name = None):
    if (camera_name is None):
        client.sio.emit('ResetLightLink')
    else:
        client.sio.emit('SetLightLink', camera_name)
	
main = Camera(main = True)


"""Sanitizing inputs to send through API"""
import numpy as np

def sanitize_vector3(vector):
  """Guarantee that a vector is a vector 3, or raise an exception

  Parameters
  ----------
  input : any
      arbitrary input parameter

  Returns
  -------
  list
      vector3 as a list [x,y,z]

  Raises
  ------
  Exception
      Failed to coerce input to a length 3 list
  """
    #if statement to check type and then convert to vector3
  if(len(vector)!=3):
    raise Exception(f'Expected a vector3 but received length {len(vector)}.')
  
  if isinstance(vector, list):
      return vector
  
  if isinstance(vector, tuple) or isinstance(vector, np.ndarray):
      return list(vector)
  else:
      raise Exception('Vector3 failed to be coerced properly')



def sanitize_color(color):
    if isinstance(color, str): #if color is string
        if(color[0] == '#'): #if color has # prefix
            return(color)
        else: #adding # prefix if not present
            color = '#' + color
            return(color)    
    
    if isinstance(color, list) or isinstance(color, tuple):#if list
        # check first if you're dealing with floats between 0 and 1
        if isinstance(color[0], float):
            # we have floats, check which range we have
            if all([(x <= 1 and x >= 0) for x in color]):
                # range 0->1, re-scale to 0->255 as ints
                color = [int(255*x) for x in color]
            elif all([(x <= 255 and x >= 0) for x in color]):
                # range 0->255, round to ints
                color = [int(np.rint(x)) for x in color]
            else:
                raise Exception('Color value ranges need to conform to [R,G,B] formatted either as 0->1 or 0->255')

        # convert list of ints to a hex string
        hex_color = '#' + ('{:02X}' * 3).format(color[0], color[1], color[2])
        return(hex_color)
    
    else:
        raise Exception('Failed to re-format color input as a hex code. Please enter either an RGB list or tuple, or a hex code')

def sanitize_float(value):
    if isinstance(value, float):
        return value
    else:
        raise Exception("Parameter needs to be passed as a float.")

def sanitize_material(material):
    if isinstance(material, str):
        return(material)
    else:
        raise Exception("Material is not properly passed in as a string. Please pass in material as a string.")

def sanitize_list(input):
    if isinstance(input,list):
        return input
    else:
        return(list(input))
    
def sanitize_string(string):
    if isinstance(string, str):
        return(string)
    else:
        raise Exception("Input is not properly passed in as a string. Please pass in input as a string.")


"""Neurons"""

from . import client
import warnings
from . import utils

## Neurons renderer
counter = 0
class Neuron:
	def __init__(self, position= [0.0,0.0,0.0], color= '#FFFFFF', size= 0.02, shape= 'sphere', material= 'lit-transparent'):
		self.create()

		# position = utils.sanitize_vector3(position)
		# self.position = position
		# client.sio.emit('SetNeuronPos', {self.id: position})

		# color = utils.sanitize_color(color)
		# self.color = color
		# client.sio.emit('SetNeuronColor',{self.id: color})

		# size = utils.sanitize_float(size)
		# self.size = size
		# client.sio.emit('SetNeuronSize',{self.id: size})

		# shape = utils.sanitize_string(shape)
		# self.shape = shape
		# client.sio.emit('SetNeuronShape',{self.id: shape})

		# material = utils.sanitize_material(material)
		# self.material = material
		# client.sio.emit('SetNeuronMaterial',{self.id: material})

	def create(self):
		"""Creates neurons
		
		Parameters
		---------- 
		none

		Examples
		>>>n1 = urchin.neurons.Neuron()
		"""
		global counter
		counter += 1
		self.id = 'n' + str(counter)
		client.sio.emit('CreateNeurons', [self.id])
		self.in_unity = True

	def delete(self):
		"""Deletes neurons
		
		Parameters
		---------- 
		references object being deleted

		Examples
		>>>n1.delete()
		"""
		client.sio.emit('DeleteNeurons', [self.id])
		self.in_unity = False

	def set_position(self, position):
		"""Set the position of neuron position in ml/ap/dv coordinates relative to the CCF (0,0,0) point
		
		Parameters
		---------- 
		position : list of three floats
			vertex positions of the neuron relative to the CCF point

		Examples
		--------
		>>>n1.set_position([2,2,2])
		"""
		if self.in_unity == False:
			raise Exception("Neuron does not exist in Unity, call create method first.")
		
		position = utils.sanitize_vector3(position)
		self.position = position
		client.sio.emit('SetNeuronPos', {self.id: position})

	def set_size(self, size):
		"""Set the size of neuron renderer
		
		Parameters
		---------- 
		size : float
			size of the neuron

		Examples
		--------
		>>>n1.set_size(0.02)
		"""
		if self.in_unity == False:
			raise Exception("Neuron does not exist in Unity, call create method first.")
		
		size = utils.sanitize_float(size)
		self.size = size
		client.sio.emit('SetNeuronSize', {self.id: size})
	
	def set_color(self, color):
		"""Set the color of neuron renderer
		
		Parameters
		---------- 
		color : string hex color
			new hex color of the neuron

		Examples
		--------
		>>>n1.set_color('#FFFFFF')
		"""
		if self.in_unity == False:
			raise Exception("Neuron does not exist in Unity, call create method first.")
		
		color = utils.sanitize_color(color)
		self.color = color
		client.sio.emit('SetNeuronColor', {self.id: color})

	def set_shape(self, shape):
		"""Set the shape of neuron renderer
		Options are
	 	 - 'sphere' (default)
	 	 - 'cube' better performance when rendering tens of thousands of neurons

		Parameters
		---------- 
		shape : string
			new shape of the neuron

		Examples
		--------
		>>>n1.set_shape('sphere')
		"""
		if self.in_unity == False:
			raise Exception("Neuron does not exist in Unity, call create method first.")
		
		self.shape = shape
		client.sio.emit('SetNeuronShape', {self.id: shape})

	def set_material(self, material):
		"""Set the material of neuron renderer
		Options are
	 	- 'lit-transparent' (default)
	 	- 'lit'
	 	- 'unlit'

		Parameters
		---------- 
		material : string
			new material of the neuron

		Examples
		--------
		>>>n1.set_material('lit-transparent')
		"""
		if self.in_unity == False:
			raise Exception("Neuron does not exist in Unity, call create method first.")
		
		self.material = material
		client.sio.emit('SetNeuronMaterial', {self.id: material})

def create(num_neurons):
	"""Create neuron objects

	Note: neurons must be created before setting other values

	Parameters
	----------
	num_neurons : int
		number of new neuron objects

	Examples
	--------
	>>> neurons = urchin.neurons.create(3)
	"""
	neuron_names = []
	for i in range(num_neurons):
		neuron_names.append(Neuron())
	return neuron_names

def delete(neurons_list):
  """Delete neuron objects

  Parameters
  ----------
  neuron_names : list of neuron objects
	list of neurons being deleted

  Examples
  --------
  >>> urchin.neurons.delete()
  """
  neurons_list = utils.sanitize_list(neurons_list)

  neurons_ids = [x.id for x in neurons_list]
  client.sio.emit('DeleteNeurons', neurons_ids)

def set_positions(neurons_list, positions_list):
	"""Set the position of neuron position in ml/ap/dv coordinates relative to the CCF (0,0,0) point

	Parameters
	----------
	neurons_list : list of neuron objects
		list of neurons being moved
	positions : list of list of three floats
		list of positions of neurons

	Examples
	--------
	>>> urchin.neurons.set_positions([n1,n2,n3], [[1,1,1],[2,2,2],[3,3,3]])
	"""
	neurons_list = utils.sanitize_list(neurons_list)
	positions_list = utils.sanitize_list(positions_list)

	neurons_pos = {}
	for i in range(len(neurons_list)):
		neuron = neurons_list[i]
		if neuron.in_unity:
			neurons_pos[neuron.id] = utils.sanitize_vector3(positions_list[i])
		else:
			warnings.warn(f"Neuron with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetNeuronPos', neurons_pos)

def set_sizes(neurons_list, sizes_list):
	"""Set neuron sizes

	Parameters
	----------
	neurons_list : list of neuron objects
		list of neurons being resized
	sizes : list of floats
		list of sizes of neurons

	Examples
	--------
	>>> urchin.neurons.set_sizes([n1,n2,n3], [0.01,0.02,0.03])
	"""
	neurons_list = utils.sanitize_list(neurons_list)
	sizes_list = utils.sanitize_list(sizes_list)

	neurons_sizes = {}
	for i in range(len(neurons_list)):
		neuron = neurons_list[i]
		if neuron.in_unity:
			neurons_sizes[neuron.id] = utils.sanitize_float(sizes_list[i])
		else:
			warnings.warn(f"Neuron with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetNeuronSize', neurons_sizes)

def set_colors(neurons_list, colors_list):
	"""Set neuron colors

	Parameters
	----------
	neurons_list : list of neuron objects
		list of neurons being recolored
	colors : list of string hex colors
		list of colors of neurons

	Examples
	--------
	>>> urchin.neurons.set_colors([n1,n2,n3], ['#FFFFFF','#000000','#FF0000'])
	"""
	neurons_list = utils.sanitize_list(neurons_list)
	colors_list = utils.sanitize_list(colors_list)

	neurons_colors = {}
	for i in range(len(neurons_list)):
		neuron = neurons_list[i]
		if neuron.in_unity:
			neurons_colors[neuron.id] = utils.sanitize_color(colors_list[i])
		else:
			warnings.warn(f"Neuron with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetNeuronColor', neurons_colors)

def set_shapes(neurons_list, shapes_list):
	"""Set neuron shapes

	Options are
	 - 'sphere' (default)
	 - 'cube' better performance when rendering tens of thousands of neurons

	Parameters
	----------
	neurons_list : list of neuron objects
		list of neurons being reshaped
	shapes : list of string
		list of shapes of neurons

	Examples
	--------
	>>> urchin.neurons.set_shapes([n1,n2,n3], ['sphere','cube','sphere'])
	"""
	neurons_list = utils.sanitize_list(neurons_list)
	shapes_list = utils.sanitize_list(shapes_list)

	neurons_shapes = {}
	for i in range(len(neurons_list)):
		neuron = neurons_list[i]
		if neuron.in_unity:
			neurons_shapes[neuron.id] = shapes_list[i]
		else:
			warnings.warn(f"Neuron with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetNeuronShape', neurons_shapes)

def set_materials(neurons_list, materials_list):
	"""Change the material used to render neurons

 	Options are
 	- 'lit-transparent' (default)
 	- 'lit'
 	- 'unlit'

	Parameters
	----------
	neurons_list : list of neuron objects
		list of neurons being rematerialized
	materials : list of string
		list of materials of neurons

	Examples
	--------
	>>> urchin.neurons.set_materials([n1,n2,n3], ['lit','unlit','lit'])
	"""
	neurons_list = utils.sanitize_list(neurons_list)
	materials_list = utils.sanitize_list(materials_list)

	neurons_materials = {}
	for i in range(len(neurons_list)):
		neuron = neurons_list[i]
		if neuron.in_unity:
			neurons_materials[neuron.id] = utils.sanitize_material(materials_list[i])
		else:
			warnings.warn(f"Neuron with id {neuron.id} does not exist in Unity, call create method first.")
	client.sio.emit('SetNeuronMaterial', neurons_materials)




"""Client for communicating with the echo server"""
import socketio
import os
import uuid

from . import camera

class bcolors:
    WARNING = '\033[93m'
    FAIL = '\033[91m'

ID = str(uuid.uuid1())[:8]

sio = socketio.Client()
@sio.event
def connect():
	print("(URN) connected to server")
	change_id(ID)

@sio.event
def disconnect():
    print("(URN) disconnected from server")

@sio.on('log')
def message(data):
	print('(Renderer) ' + data)

@sio.on('log-warning')
def message(data):
	print('(Renderer) ' + bcolors.WARNING + data)

@sio.on('log-error')
def message(data):
	print('(Renderer) ' + bcolors.FAIL + data)

@sio.on('ReceiveCameraImgMeta')
def receive_camera_img_meta(data):
	camera.receive_camera_img_meta(data)
	
@sio.on('ReceiveCameraImg')
def receive_camera_img(data):
  camera.receive_camera_img(data)
def connected():
	return sio.connected

def close():
	"""Disconnect from the echo server
	"""
	sio.disconnect()

def change_id(newID):
	"""Change the ID used to connect to the echo server

	Parameters
	----------
	newID : string
		New ID to connect with
	"""
	sio.emit('ID',[newID,"send"])
	print(f'Login sent with ID: {newID}, copy this ID into the renderer to connect.')


""" Allen CCF 3D Models"""

from .. import client

def load_beryl():
	"""Load all beryl areas and set visibility to True

	NOTE: One of the load functions OR set_area_visibility must be called
	before you set the color/alpha/etc of brain regions.
	"""
	client.sio.emit('LoadDefaultAreas', 'beryl')

def load_cosmos():
	"""Load all cosmos areas and set visibility to True

	NOTE: One of the load functions OR set_area_visibility must be called
	before you set the color/alpha/etc of brain regions.
	"""
	client.sio.emit('LoadDefaultAreas', 'cosmos')

def set_visibility(area_visibilities):
	"""Set visibility of CCF area models

	**Note:** you can append a "-lh" or "-rh" suffix to any area acronym or ID to control the visibility of just one-half of the model. This can be used in all of the set_area_* functions.

	Parameters
	----------
	area_visibilities : dict {string : bool}
		dictionary of area IDs or acronyms and visibility values

	Examples
	--------
	>>> urn.set_visibility({'grey':True})
	>>> urn.set_visibility({8:True})
	>>> urn.set_visibility({'VISp-l':True})
	"""
	client.sio.emit('SetAreaVisibility', area_visibilities)

def set_color(area_colors):
	"""Set color of CCF area models.

	Parameters
	----------
	area_colors : dict {string: string}
		Keys are area IDs or acronyms, Values are hex colors

	Examples
	--------
	>>> urn.set_color( {'grey':"#FFFFFF"})
	>>> urn.set_color({8:"#FFFFFF"})
	>>> urn.set_color({"VISp-l":"#00000080"})
	"""
	client.sio.emit('SetAreaColors', area_colors)

def set_intensity(area_intensities):
	"""Set color of CCF area models using colormap.

	Parameters
	----------
	area_intensities : dict {string: float}
		keys are area IDs or acronyms, values are hex colors

	Examples
	--------
	>>> urn.set_intensity( {'grey':1.0})
	>>> urn.set_intensity({8:1.0})
	"""
	client.sio.emit('SetAreaIntensity', area_intensities)

def set_colormap(colormap_name):
	"""Set colormap used for CCF area intensity mapping


	Options are
	 - cool (default, teal 0 -> magenta 1)
	 - grey (black 0 -> white 1)
	 - grey-green (grey 0, light 1/255 -> dark 1)
	 - grey-purple (grey 0, light 1/255 -> dark 1)
	 - grey-red (grey 0, light 1/255 -> dark 1)
	 - grey-rainbow (grey 0, rainbow colors from 1/255 -> 1)

	Parameters
	----------
	colormap_name : string
		colormap name
	"""
	client.sio.emit('SetAreaColormap', colormap_name)

def set_alpha(area_alpha):
	"""Set transparency of CCF area models. 

	Parameters
	----------
	area_alpha : dict {string: float}
		keys are area IDs or acronyms, values are percent transparency

	Examples
	--------
	>>> urn.set_alpha( {'grey':0.5})
	>>> urn.set_alpha({8:0.5})
	"""
	client.sio.emit('SetAreaAlpha', area_alpha)

def set_material(area_materials):
	"""Set material of CCF area models.

	Material options are
	 - 'opaque-lit' or 'default'
	 - 'opaque-unlit'
	 - 'transparent-lit'
	 - 'transparent-unlit'

	Parameters
	----------
	area_materials : dict {string: string}
		keys are area IDs or acronyms, values are material options
	"""
	client.sio.emit('SetAreaMaterial', area_materials)

def set_data(area_data):
	"""Set the data array for each CCF area model

	Data arrays work the same as the set_area_intensity() function but are controlled by the area_index value, which can be set in the renderer or through the API.

	Parameters
	----------
	area_data : dict {string: float list}
		keys area IDs or acronyms, values are a list of floats
	"""
	client.sio.emit('SetAreaData', area_data)

def set_data_index(area_index):
	"""Set the data index for the CCF area models

	Parameters
	----------
	area_index : int
		data index
	"""
	client.sio.emit('SetAreaIndex', area_index)


"""Volumetric datasets (x*y*z matrix)"""

from . import client
import numpy as np

def create(volume_name):
	"""Create an empty volume data matrix in the renderer.

	Note: you must call create_volume and set_volume_colormap before setting data.

	Parameters
	----------
	volume_name : string
		volume name

	Examples
	--------
	>>> urn.create('histology')
	"""
	client.sio.emit('CreateVolume', volume_name)

def delete(volume_name):
	"""Delete a volume data matrix.

	Parameters
	----------
	volume_name : string
		volume name

	Examples
	--------
	>>> urn.delete('histology')
	"""
	client.sio.emit('DeleteVolume', volume_name)

	
def set_colormap(volume_name, colormap):
	"""Set the colormap for a volume, maximum of 254 values

	Note: index 255 is reserved by the renderer for transparency.

	Parameters
	----------
	volumeName : string
		volume name
	volumeData : list of string
		list of hex colors, values can be RGB or RGBA
		
	Examples
	--------
	>>> urn.set_colormap('histology',['#800000','#FF0000'])
	"""
	data_packet = colormap.copy()
	data_packet.insert(0,volume_name)
	client.sio.emit('SetVolumeColormap', data_packet)

def set_visibility(volume_name, volume_visibility):
	"""Change the visibility of a volume

	Parameters
	----------
	volumeName : string
		Volume name
	volumeVisibility : bool
		New visibility setting
		
	Examples
	--------
	>>> urn.set_visibility('histology', True)
	"""
	client.sio.emit('SetVolumeVisibility', [volume_name, volume_visibility])

def set_data(volume_name, volumeData):
	"""Set the data for a volume using uint8 values from 0->254 (255 is reserved for transparent).
	Sending your data as any type other than np.uint8 is potentially unsafe.
	Data will be remapped in the renderer according to the active colormap.
	nan values will be set to transparent.

	Note: this function slices the data by depth and sends data slice-by-slice, it may take a long time to run.

	Parameters
	----------
	volumeData : [string, numpy matrix]
		Name of the volume and the volume itself. Volume should be [528, 320, 456] with +x = posterior, +y = ventral, +z = right
	"""
	localData = volumeData.copy()
	localData[np.isnan(localData)] = 255
	localData = localData.astype(np.uint8)

	ndepth = localData.shape[2]

	for di in np.arange(0,ndepth):
		if di == (ndepth-1):
			set_slice_data(volume_name, int(ndepth-1-di), localData[:,:,di].flatten('F').tobytes(), True)
		else:
			set_slice_data(volume_name, int(ndepth-1-di), localData[:,:,di].flatten('F').tobytes())

def set_slice_data(volume_name, slice, volume_bytes, override_gpu_apply = False):
	"""Set a single slice of data in a volume

	Parameters
	----------
	volumeName : string
		name of volume
	slice : int
		depth slice in volume (on lr dimension)
	volume_bytes : bytes
		flattened bytes array from slice
	override_gpu_apply : bool, optional
		immediately set the GPU texture data after this slice is sent, by default False
	"""
	client.sio.emit('SetVolumeDataMeta', [volume_name, slice, override_gpu_apply])
	client.sio.emit('SetVolumeData', volume_bytes)

######################
# VOLUMES: Allen CCF #
######################

# special case of the regular volume visibility function
def set_visibility_allen(allenVisibility):
	"""Set the visibility of the Allen CCF annotation volume

	Parameters
	----------
	allenData : bool
		visibility of Allen CCF volume
	"""
	set_visibility('allen',allenVisibility)



# From https://github.com/mwaskom/seaborn/blob/master/seaborn/colors/xkcd_rgb.py

xkcd_rgb = {'acid green': '#8ffe09',
            'adobe': '#bd6c48',
            'algae': '#54ac68',
            'algae green': '#21c36f',
            'almost black': '#070d0d',
            'amber': '#feb308',
            'amethyst': '#9b5fc0',
            'apple': '#6ecb3c',
            'apple green': '#76cd26',
            'apricot': '#ffb16d',
            'aqua': '#13eac9',
            'aqua blue': '#02d8e9',
            'aqua green': '#12e193',
            'aqua marine': '#2ee8bb',
            'aquamarine': '#04d8b2',
            'army green': '#4b5d16',
            'asparagus': '#77ab56',
            'aubergine': '#3d0734',
            'auburn': '#9a3001',
            'avocado': '#90b134',
            'avocado green': '#87a922',
            'azul': '#1d5dec',
            'azure': '#069af3',
            'baby blue': '#a2cffe',
            'baby green': '#8cff9e',
            'baby pink': '#ffb7ce',
            'baby poo': '#ab9004',
            'baby poop': '#937c00',
            'baby poop green': '#8f9805',
            'baby puke green': '#b6c406',
            'baby purple': '#ca9bf7',
            'baby shit brown': '#ad900d',
            'baby shit green': '#889717',
            'banana': '#ffff7e',
            'banana yellow': '#fafe4b',
            'barbie pink': '#fe46a5',
            'barf green': '#94ac02',
            'barney': '#ac1db8',
            'barney purple': '#a00498',
            'battleship grey': '#6b7c85',
            'beige': '#e6daa6',
            'berry': '#990f4b',
            'bile': '#b5c306',
            'black': '#000000',
            'bland': '#afa88b',
            'blood': '#770001',
            'blood orange': '#fe4b03',
            'blood red': '#980002',
            'blue': '#0343df',
            'blue blue': '#2242c7',
            'blue green': '#137e6d',
            'blue grey': '#607c8e',
            'blue purple': '#5729ce',
            'blue violet': '#5d06e9',
            'blue with a hint of purple': '#533cc6',
            'blue/green': '#0f9b8e',
            'blue/grey': '#758da3',
            'blue/purple': '#5a06ef',
            'blueberry': '#464196',
            'bluegreen': '#017a79',
            'bluegrey': '#85a3b2',
            'bluey green': '#2bb179',
            'bluey grey': '#89a0b0',
            'bluey purple': '#6241c7',
            'bluish': '#2976bb',
            'bluish green': '#10a674',
            'bluish grey': '#748b97',
            'bluish purple': '#703be7',
            'blurple': '#5539cc',
            'blush': '#f29e8e',
            'blush pink': '#fe828c',
            'booger': '#9bb53c',
            'booger green': '#96b403',
            'bordeaux': '#7b002c',
            'boring green': '#63b365',
            'bottle green': '#044a05',
            'brick': '#a03623',
            'brick orange': '#c14a09',
            'brick red': '#8f1402',
            'bright aqua': '#0bf9ea',
            'bright blue': '#0165fc',
            'bright cyan': '#41fdfe',
            'bright green': '#01ff07',
            'bright lavender': '#c760ff',
            'bright light blue': '#26f7fd',
            'bright light green': '#2dfe54',
            'bright lilac': '#c95efb',
            'bright lime': '#87fd05',
            'bright lime green': '#65fe08',
            'bright magenta': '#ff08e8',
            'bright olive': '#9cbb04',
            'bright orange': '#ff5b00',
            'bright pink': '#fe01b1',
            'bright purple': '#be03fd',
            'bright red': '#ff000d',
            'bright sea green': '#05ffa6',
            'bright sky blue': '#02ccfe',
            'bright teal': '#01f9c6',
            'bright turquoise': '#0ffef9',
            'bright violet': '#ad0afd',
            'bright yellow': '#fffd01',
            'bright yellow green': '#9dff00',
            'british racing green': '#05480d',
            'bronze': '#a87900',
            'brown': '#653700',
            'brown green': '#706c11',
            'brown grey': '#8d8468',
            'brown orange': '#b96902',
            'brown red': '#922b05',
            'brown yellow': '#b29705',
            'brownish': '#9c6d57',
            'brownish green': '#6a6e09',
            'brownish grey': '#86775f',
            'brownish orange': '#cb7723',
            'brownish pink': '#c27e79',
            'brownish purple': '#76424e',
            'brownish red': '#9e3623',
            'brownish yellow': '#c9b003',
            'browny green': '#6f6c0a',
            'browny orange': '#ca6b02',
            'bruise': '#7e4071',
            'bubble gum pink': '#ff69af',
            'bubblegum': '#ff6cb5',
            'bubblegum pink': '#fe83cc',
            'buff': '#fef69e',
            'burgundy': '#610023',
            'burnt orange': '#c04e01',
            'burnt red': '#9f2305',
            'burnt siena': '#b75203',
            'burnt sienna': '#b04e0f',
            'burnt umber': '#a0450e',
            'burnt yellow': '#d5ab09',
            'burple': '#6832e3',
            'butter': '#ffff81',
            'butter yellow': '#fffd74',
            'butterscotch': '#fdb147',
            'cadet blue': '#4e7496',
            'camel': '#c69f59',
            'camo': '#7f8f4e',
            'camo green': '#526525',
            'camouflage green': '#4b6113',
            'canary': '#fdff63',
            'canary yellow': '#fffe40',
            'candy pink': '#ff63e9',
            'caramel': '#af6f09',
            'carmine': '#9d0216',
            'carnation': '#fd798f',
            'carnation pink': '#ff7fa7',
            'carolina blue': '#8ab8fe',
            'celadon': '#befdb7',
            'celery': '#c1fd95',
            'cement': '#a5a391',
            'cerise': '#de0c62',
            'cerulean': '#0485d1',
            'cerulean blue': '#056eee',
            'charcoal': '#343837',
            'charcoal grey': '#3c4142',
            'chartreuse': '#c1f80a',
            'cherry': '#cf0234',
            'cherry red': '#f7022a',
            'chestnut': '#742802',
            'chocolate': '#3d1c02',
            'chocolate brown': '#411900',
            'cinnamon': '#ac4f06',
            'claret': '#680018',
            'clay': '#b66a50',
            'clay brown': '#b2713d',
            'clear blue': '#247afd',
            'cloudy blue': '#acc2d9',
            'cobalt': '#1e488f',
            'cobalt blue': '#030aa7',
            'cocoa': '#875f42',
            'coffee': '#a6814c',
            'cool blue': '#4984b8',
            'cool green': '#33b864',
            'cool grey': '#95a3a6',
            'copper': '#b66325',
            'coral': '#fc5a50',
            'coral pink': '#ff6163',
            'cornflower': '#6a79f7',
            'cornflower blue': '#5170d7',
            'cranberry': '#9e003a',
            'cream': '#ffffc2',
            'creme': '#ffffb6',
            'crimson': '#8c000f',
            'custard': '#fffd78',
            'cyan': '#00ffff',
            'dandelion': '#fedf08',
            'dark': '#1b2431',
            'dark aqua': '#05696b',
            'dark aquamarine': '#017371',
            'dark beige': '#ac9362',
            'dark blue': '#00035b',
            'dark blue green': '#005249',
            'dark blue grey': '#1f3b4d',
            'dark brown': '#341c02',
            'dark coral': '#cf524e',
            'dark cream': '#fff39a',
            'dark cyan': '#0a888a',
            'dark forest green': '#002d04',
            'dark fuchsia': '#9d0759',
            'dark gold': '#b59410',
            'dark grass green': '#388004',
            'dark green': '#033500',
            'dark green blue': '#1f6357',
            'dark grey': '#363737',
            'dark grey blue': '#29465b',
            'dark hot pink': '#d90166',
            'dark indigo': '#1f0954',
            'dark khaki': '#9b8f55',
            'dark lavender': '#856798',
            'dark lilac': '#9c6da5',
            'dark lime': '#84b701',
            'dark lime green': '#7ebd01',
            'dark magenta': '#960056',
            'dark maroon': '#3c0008',
            'dark mauve': '#874c62',
            'dark mint': '#48c072',
            'dark mint green': '#20c073',
            'dark mustard': '#a88905',
            'dark navy': '#000435',
            'dark navy blue': '#00022e',
            'dark olive': '#373e02',
            'dark olive green': '#3c4d03',
            'dark orange': '#c65102',
            'dark pastel green': '#56ae57',
            'dark peach': '#de7e5d',
            'dark periwinkle': '#665fd1',
            'dark pink': '#cb416b',
            'dark plum': '#3f012c',
            'dark purple': '#35063e',
            'dark red': '#840000',
            'dark rose': '#b5485d',
            'dark royal blue': '#02066f',
            'dark sage': '#598556',
            'dark salmon': '#c85a53',
            'dark sand': '#a88f59',
            'dark sea green': '#11875d',
            'dark seafoam': '#1fb57a',
            'dark seafoam green': '#3eaf76',
            'dark sky blue': '#448ee4',
            'dark slate blue': '#214761',
            'dark tan': '#af884a',
            'dark taupe': '#7f684e',
            'dark teal': '#014d4e',
            'dark turquoise': '#045c5a',
            'dark violet': '#34013f',
            'dark yellow': '#d5b60a',
            'dark yellow green': '#728f02',
            'darkblue': '#030764',
            'darkgreen': '#054907',
            'darkish blue': '#014182',
            'darkish green': '#287c37',
            'darkish pink': '#da467d',
            'darkish purple': '#751973',
            'darkish red': '#a90308',
            'deep aqua': '#08787f',
            'deep blue': '#040273',
            'deep brown': '#410200',
            'deep green': '#02590f',
            'deep lavender': '#8d5eb7',
            'deep lilac': '#966ebd',
            'deep magenta': '#a0025c',
            'deep orange': '#dc4d01',
            'deep pink': '#cb0162',
            'deep purple': '#36013f',
            'deep red': '#9a0200',
            'deep rose': '#c74767',
            'deep sea blue': '#015482',
            'deep sky blue': '#0d75f8',
            'deep teal': '#00555a',
            'deep turquoise': '#017374',
            'deep violet': '#490648',
            'denim': '#3b638c',
            'denim blue': '#3b5b92',
            'desert': '#ccad60',
            'diarrhea': '#9f8303',
            'dirt': '#8a6e45',
            'dirt brown': '#836539',
            'dirty blue': '#3f829d',
            'dirty green': '#667e2c',
            'dirty orange': '#c87606',
            'dirty pink': '#ca7b80',
            'dirty purple': '#734a65',
            'dirty yellow': '#cdc50a',
            'dodger blue': '#3e82fc',
            'drab': '#828344',
            'drab green': '#749551',
            'dried blood': '#4b0101',
            'duck egg blue': '#c3fbf4',
            'dull blue': '#49759c',
            'dull brown': '#876e4b',
            'dull green': '#74a662',
            'dull orange': '#d8863b',
            'dull pink': '#d5869d',
            'dull purple': '#84597e',
            'dull red': '#bb3f3f',
            'dull teal': '#5f9e8f',
            'dull yellow': '#eedc5b',
            'dusk': '#4e5481',
            'dusk blue': '#26538d',
            'dusky blue': '#475f94',
            'dusky pink': '#cc7a8b',
            'dusky purple': '#895b7b',
            'dusky rose': '#ba6873',
            'dust': '#b2996e',
            'dusty blue': '#5a86ad',
            'dusty green': '#76a973',
            'dusty lavender': '#ac86a8',
            'dusty orange': '#f0833a',
            'dusty pink': '#d58a94',
            'dusty purple': '#825f87',
            'dusty red': '#b9484e',
            'dusty rose': '#c0737a',
            'dusty teal': '#4c9085',
            'earth': '#a2653e',
            'easter green': '#8cfd7e',
            'easter purple': '#c071fe',
            'ecru': '#feffca',
            'egg shell': '#fffcc4',
            'eggplant': '#380835',
            'eggplant purple': '#430541',
            'eggshell': '#ffffd4',
            'eggshell blue': '#c4fff7',
            'electric blue': '#0652ff',
            'electric green': '#21fc0d',
            'electric lime': '#a8ff04',
            'electric pink': '#ff0490',
            'electric purple': '#aa23ff',
            'emerald': '#01a049',
            'emerald green': '#028f1e',
            'evergreen': '#05472a',
            'faded blue': '#658cbb',
            'faded green': '#7bb274',
            'faded orange': '#f0944d',
            'faded pink': '#de9dac',
            'faded purple': '#916e99',
            'faded red': '#d3494e',
            'faded yellow': '#feff7f',
            'fawn': '#cfaf7b',
            'fern': '#63a950',
            'fern green': '#548d44',
            'fire engine red': '#fe0002',
            'flat blue': '#3c73a8',
            'flat green': '#699d4c',
            'fluorescent green': '#08ff08',
            'fluro green': '#0aff02',
            'foam green': '#90fda9',
            'forest': '#0b5509',
            'forest green': '#06470c',
            'forrest green': '#154406',
            'french blue': '#436bad',
            'fresh green': '#69d84f',
            'frog green': '#58bc08',
            'fuchsia': '#ed0dd9',
            'gold': '#dbb40c',
            'golden': '#f5bf03',
            'golden brown': '#b27a01',
            'golden rod': '#f9bc08',
            'golden yellow': '#fec615',
            'goldenrod': '#fac205',
            'grape': '#6c3461',
            'grape purple': '#5d1451',
            'grapefruit': '#fd5956',
            'grass': '#5cac2d',
            'grass green': '#3f9b0b',
            'grassy green': '#419c03',
            'green': '#15b01a',
            'green apple': '#5edc1f',
            'green blue': '#06b48b',
            'green brown': '#544e03',
            'green grey': '#77926f',
            'green teal': '#0cb577',
            'green yellow': '#c9ff27',
            'green/blue': '#01c08d',
            'green/yellow': '#b5ce08',
            'greenblue': '#23c48b',
            'greenish': '#40a368',
            'greenish beige': '#c9d179',
            'greenish blue': '#0b8b87',
            'greenish brown': '#696112',
            'greenish cyan': '#2afeb7',
            'greenish grey': '#96ae8d',
            'greenish tan': '#bccb7a',
            'greenish teal': '#32bf84',
            'greenish turquoise': '#00fbb0',
            'greenish yellow': '#cdfd02',
            'greeny blue': '#42b395',
            'greeny brown': '#696006',
            'greeny grey': '#7ea07a',
            'greeny yellow': '#c6f808',
            'grey': '#929591',
            'grey blue': '#6b8ba4',
            'grey brown': '#7f7053',
            'grey green': '#789b73',
            'grey pink': '#c3909b',
            'grey purple': '#826d8c',
            'grey teal': '#5e9b8a',
            'grey/blue': '#647d8e',
            'grey/green': '#86a17d',
            'greyblue': '#77a1b5',
            'greyish': '#a8a495',
            'greyish blue': '#5e819d',
            'greyish brown': '#7a6a4f',
            'greyish green': '#82a67d',
            'greyish pink': '#c88d94',
            'greyish purple': '#887191',
            'greyish teal': '#719f91',
            'gross green': '#a0bf16',
            'gunmetal': '#536267',
            'hazel': '#8e7618',
            'heather': '#a484ac',
            'heliotrope': '#d94ff5',
            'highlighter green': '#1bfc06',
            'hospital green': '#9be5aa',
            'hot green': '#25ff29',
            'hot magenta': '#f504c9',
            'hot pink': '#ff028d',
            'hot purple': '#cb00f5',
            'hunter green': '#0b4008',
            'ice': '#d6fffa',
            'ice blue': '#d7fffe',
            'icky green': '#8fae22',
            'indian red': '#850e04',
            'indigo': '#380282',
            'indigo blue': '#3a18b1',
            'iris': '#6258c4',
            'irish green': '#019529',
            'ivory': '#ffffcb',
            'jade': '#1fa774',
            'jade green': '#2baf6a',
            'jungle green': '#048243',
            'kelley green': '#009337',
            'kelly green': '#02ab2e',
            'kermit green': '#5cb200',
            'key lime': '#aeff6e',
            'khaki': '#aaa662',
            'khaki green': '#728639',
            'kiwi': '#9cef43',
            'kiwi green': '#8ee53f',
            'lavender': '#c79fef',
            'lavender blue': '#8b88f8',
            'lavender pink': '#dd85d7',
            'lawn green': '#4da409',
            'leaf': '#71aa34',
            'leaf green': '#5ca904',
            'leafy green': '#51b73b',
            'leather': '#ac7434',
            'lemon': '#fdff52',
            'lemon green': '#adf802',
            'lemon lime': '#bffe28',
            'lemon yellow': '#fdff38',
            'lichen': '#8fb67b',
            'light aqua': '#8cffdb',
            'light aquamarine': '#7bfdc7',
            'light beige': '#fffeb6',
            'light blue': '#95d0fc',
            'light blue green': '#7efbb3',
            'light blue grey': '#b7c9e2',
            'light bluish green': '#76fda8',
            'light bright green': '#53fe5c',
            'light brown': '#ad8150',
            'light burgundy': '#a8415b',
            'light cyan': '#acfffc',
            'light eggplant': '#894585',
            'light forest green': '#4f9153',
            'light gold': '#fddc5c',
            'light grass green': '#9af764',
            'light green': '#96f97b',
            'light green blue': '#56fca2',
            'light greenish blue': '#63f7b4',
            'light grey': '#d8dcd6',
            'light grey blue': '#9dbcd4',
            'light grey green': '#b7e1a1',
            'light indigo': '#6d5acf',
            'light khaki': '#e6f2a2',
            'light lavendar': '#efc0fe',
            'light lavender': '#dfc5fe',
            'light light blue': '#cafffb',
            'light light green': '#c8ffb0',
            'light lilac': '#edc8ff',
            'light lime': '#aefd6c',
            'light lime green': '#b9ff66',
            'light magenta': '#fa5ff7',
            'light maroon': '#a24857',
            'light mauve': '#c292a1',
            'light mint': '#b6ffbb',
            'light mint green': '#a6fbb2',
            'light moss green': '#a6c875',
            'light mustard': '#f7d560',
            'light navy': '#155084',
            'light navy blue': '#2e5a88',
            'light neon green': '#4efd54',
            'light olive': '#acbf69',
            'light olive green': '#a4be5c',
            'light orange': '#fdaa48',
            'light pastel green': '#b2fba5',
            'light pea green': '#c4fe82',
            'light peach': '#ffd8b1',
            'light periwinkle': '#c1c6fc',
            'light pink': '#ffd1df',
            'light plum': '#9d5783',
            'light purple': '#bf77f6',
            'light red': '#ff474c',
            'light rose': '#ffc5cb',
            'light royal blue': '#3a2efe',
            'light sage': '#bcecac',
            'light salmon': '#fea993',
            'light sea green': '#98f6b0',
            'light seafoam': '#a0febf',
            'light seafoam green': '#a7ffb5',
            'light sky blue': '#c6fcff',
            'light tan': '#fbeeac',
            'light teal': '#90e4c1',
            'light turquoise': '#7ef4cc',
            'light urple': '#b36ff6',
            'light violet': '#d6b4fc',
            'light yellow': '#fffe7a',
            'light yellow green': '#ccfd7f',
            'light yellowish green': '#c2ff89',
            'lightblue': '#7bc8f6',
            'lighter green': '#75fd63',
            'lighter purple': '#a55af4',
            'lightgreen': '#76ff7b',
            'lightish blue': '#3d7afd',
            'lightish green': '#61e160',
            'lightish purple': '#a552e6',
            'lightish red': '#fe2f4a',
            'lilac': '#cea2fd',
            'liliac': '#c48efd',
            'lime': '#aaff32',
            'lime green': '#89fe05',
            'lime yellow': '#d0fe1d',
            'lipstick': '#d5174e',
            'lipstick red': '#c0022f',
            'macaroni and cheese': '#efb435',
            'magenta': '#c20078',
            'mahogany': '#4a0100',
            'maize': '#f4d054',
            'mango': '#ffa62b',
            'manilla': '#fffa86',
            'marigold': '#fcc006',
            'marine': '#042e60',
            'marine blue': '#01386a',
            'maroon': '#650021',
            'mauve': '#ae7181',
            'medium blue': '#2c6fbb',
            'medium brown': '#7f5112',
            'medium green': '#39ad48',
            'medium grey': '#7d7f7c',
            'medium pink': '#f36196',
            'medium purple': '#9e43a2',
            'melon': '#ff7855',
            'merlot': '#730039',
            'metallic blue': '#4f738e',
            'mid blue': '#276ab3',
            'mid green': '#50a747',
            'midnight': '#03012d',
            'midnight blue': '#020035',
            'midnight purple': '#280137',
            'military green': '#667c3e',
            'milk chocolate': '#7f4e1e',
            'mint': '#9ffeb0',
            'mint green': '#8fff9f',
            'minty green': '#0bf77d',
            'mocha': '#9d7651',
            'moss': '#769958',
            'moss green': '#658b38',
            'mossy green': '#638b27',
            'mud': '#735c12',
            'mud brown': '#60460f',
            'mud green': '#606602',
            'muddy brown': '#886806',
            'muddy green': '#657432',
            'muddy yellow': '#bfac05',
            'mulberry': '#920a4e',
            'murky green': '#6c7a0e',
            'mushroom': '#ba9e88',
            'mustard': '#ceb301',
            'mustard brown': '#ac7e04',
            'mustard green': '#a8b504',
            'mustard yellow': '#d2bd0a',
            'muted blue': '#3b719f',
            'muted green': '#5fa052',
            'muted pink': '#d1768f',
            'muted purple': '#805b87',
            'nasty green': '#70b23f',
            'navy': '#01153e',
            'navy blue': '#001146',
            'navy green': '#35530a',
            'neon blue': '#04d9ff',
            'neon green': '#0cff0c',
            'neon pink': '#fe019a',
            'neon purple': '#bc13fe',
            'neon red': '#ff073a',
            'neon yellow': '#cfff04',
            'nice blue': '#107ab0',
            'night blue': '#040348',
            'ocean': '#017b92',
            'ocean blue': '#03719c',
            'ocean green': '#3d9973',
            'ocher': '#bf9b0c',
            'ochre': '#bf9005',
            'ocre': '#c69c04',
            'off blue': '#5684ae',
            'off green': '#6ba353',
            'off white': '#ffffe4',
            'off yellow': '#f1f33f',
            'old pink': '#c77986',
            'old rose': '#c87f89',
            'olive': '#6e750e',
            'olive brown': '#645403',
            'olive drab': '#6f7632',
            'olive green': '#677a04',
            'olive yellow': '#c2b709',
            'orange': '#f97306',
            'orange brown': '#be6400',
            'orange pink': '#ff6f52',
            'orange red': '#fd411e',
            'orange yellow': '#ffad01',
            'orangeish': '#fd8d49',
            'orangered': '#fe420f',
            'orangey brown': '#b16002',
            'orangey red': '#fa4224',
            'orangey yellow': '#fdb915',
            'orangish': '#fc824a',
            'orangish brown': '#b25f03',
            'orangish red': '#f43605',
            'orchid': '#c875c4',
            'pale': '#fff9d0',
            'pale aqua': '#b8ffeb',
            'pale blue': '#d0fefe',
            'pale brown': '#b1916e',
            'pale cyan': '#b7fffa',
            'pale gold': '#fdde6c',
            'pale green': '#c7fdb5',
            'pale grey': '#fdfdfe',
            'pale lavender': '#eecffe',
            'pale light green': '#b1fc99',
            'pale lilac': '#e4cbff',
            'pale lime': '#befd73',
            'pale lime green': '#b1ff65',
            'pale magenta': '#d767ad',
            'pale mauve': '#fed0fc',
            'pale olive': '#b9cc81',
            'pale olive green': '#b1d27b',
            'pale orange': '#ffa756',
            'pale peach': '#ffe5ad',
            'pale pink': '#ffcfdc',
            'pale purple': '#b790d4',
            'pale red': '#d9544d',
            'pale rose': '#fdc1c5',
            'pale salmon': '#ffb19a',
            'pale sky blue': '#bdf6fe',
            'pale teal': '#82cbb2',
            'pale turquoise': '#a5fbd5',
            'pale violet': '#ceaefa',
            'pale yellow': '#ffff84',
            'parchment': '#fefcaf',
            'pastel blue': '#a2bffe',
            'pastel green': '#b0ff9d',
            'pastel orange': '#ff964f',
            'pastel pink': '#ffbacd',
            'pastel purple': '#caa0ff',
            'pastel red': '#db5856',
            'pastel yellow': '#fffe71',
            'pea': '#a4bf20',
            'pea green': '#8eab12',
            'pea soup': '#929901',
            'pea soup green': '#94a617',
            'peach': '#ffb07c',
            'peachy pink': '#ff9a8a',
            'peacock blue': '#016795',
            'pear': '#cbf85f',
            'periwinkle': '#8e82fe',
            'periwinkle blue': '#8f99fb',
            'perrywinkle': '#8f8ce7',
            'petrol': '#005f6a',
            'pig pink': '#e78ea5',
            'pine': '#2b5d34',
            'pine green': '#0a481e',
            'pink': '#ff81c0',
            'pink purple': '#db4bda',
            'pink red': '#f5054f',
            'pink/purple': '#ef1de7',
            'pinkish': '#d46a7e',
            'pinkish brown': '#b17261',
            'pinkish grey': '#c8aca9',
            'pinkish orange': '#ff724c',
            'pinkish purple': '#d648d7',
            'pinkish red': '#f10c45',
            'pinkish tan': '#d99b82',
            'pinky': '#fc86aa',
            'pinky purple': '#c94cbe',
            'pinky red': '#fc2647',
            'piss yellow': '#ddd618',
            'pistachio': '#c0fa8b',
            'plum': '#580f41',
            'plum purple': '#4e0550',
            'poison green': '#40fd14',
            'poo': '#8f7303',
            'poo brown': '#885f01',
            'poop': '#7f5e00',
            'poop brown': '#7a5901',
            'poop green': '#6f7c00',
            'powder blue': '#b1d1fc',
            'powder pink': '#ffb2d0',
            'primary blue': '#0804f9',
            'prussian blue': '#004577',
            'puce': '#a57e52',
            'puke': '#a5a502',
            'puke brown': '#947706',
            'puke green': '#9aae07',
            'puke yellow': '#c2be0e',
            'pumpkin': '#e17701',
            'pumpkin orange': '#fb7d07',
            'pure blue': '#0203e2',
            'purple': '#7e1e9c',
            'purple blue': '#632de9',
            'purple brown': '#673a3f',
            'purple grey': '#866f85',
            'purple pink': '#e03fd8',
            'purple red': '#990147',
            'purple/blue': '#5d21d0',
            'purple/pink': '#d725de',
            'purpleish': '#98568d',
            'purpleish blue': '#6140ef',
            'purpleish pink': '#df4ec8',
            'purpley': '#8756e4',
            'purpley blue': '#5f34e7',
            'purpley grey': '#947e94',
            'purpley pink': '#c83cb9',
            'purplish': '#94568c',
            'purplish blue': '#601ef9',
            'purplish brown': '#6b4247',
            'purplish grey': '#7a687f',
            'purplish pink': '#ce5dae',
            'purplish red': '#b0054b',
            'purply': '#983fb2',
            'purply blue': '#661aee',
            'purply pink': '#f075e6',
            'putty': '#beae8a',
            'racing green': '#014600',
            'radioactive green': '#2cfa1f',
            'raspberry': '#b00149',
            'raw sienna': '#9a6200',
            'raw umber': '#a75e09',
            'really light blue': '#d4ffff',
            'red': '#e50000',
            'red brown': '#8b2e16',
            'red orange': '#fd3c06',
            'red pink': '#fa2a55',
            'red purple': '#820747',
            'red violet': '#9e0168',
            'red wine': '#8c0034',
            'reddish': '#c44240',
            'reddish brown': '#7f2b0a',
            'reddish grey': '#997570',
            'reddish orange': '#f8481c',
            'reddish pink': '#fe2c54',
            'reddish purple': '#910951',
            'reddy brown': '#6e1005',
            'rich blue': '#021bf9',
            'rich purple': '#720058',
            'robin egg blue': '#8af1fe',
            "robin's egg": '#6dedfd',
            "robin's egg blue": '#98eff9',
            'rosa': '#fe86a4',
            'rose': '#cf6275',
            'rose pink': '#f7879a',
            'rose red': '#be013c',
            'rosy pink': '#f6688e',
            'rouge': '#ab1239',
            'royal': '#0c1793',
            'royal blue': '#0504aa',
            'royal purple': '#4b006e',
            'ruby': '#ca0147',
            'russet': '#a13905',
            'rust': '#a83c09',
            'rust brown': '#8b3103',
            'rust orange': '#c45508',
            'rust red': '#aa2704',
            'rusty orange': '#cd5909',
            'rusty red': '#af2f0d',
            'saffron': '#feb209',
            'sage': '#87ae73',
            'sage green': '#88b378',
            'salmon': '#ff796c',
            'salmon pink': '#fe7b7c',
            'sand': '#e2ca76',
            'sand brown': '#cba560',
            'sand yellow': '#fce166',
            'sandstone': '#c9ae74',
            'sandy': '#f1da7a',
            'sandy brown': '#c4a661',
            'sandy yellow': '#fdee73',
            'sap green': '#5c8b15',
            'sapphire': '#2138ab',
            'scarlet': '#be0119',
            'sea': '#3c9992',
            'sea blue': '#047495',
            'sea green': '#53fca1',
            'seafoam': '#80f9ad',
            'seafoam blue': '#78d1b6',
            'seafoam green': '#7af9ab',
            'seaweed': '#18d17b',
            'seaweed green': '#35ad6b',
            'sepia': '#985e2b',
            'shamrock': '#01b44c',
            'shamrock green': '#02c14d',
            'shit': '#7f5f00',
            'shit brown': '#7b5804',
            'shit green': '#758000',
            'shocking pink': '#fe02a2',
            'sick green': '#9db92c',
            'sickly green': '#94b21c',
            'sickly yellow': '#d0e429',
            'sienna': '#a9561e',
            'silver': '#c5c9c7',
            'sky': '#82cafc',
            'sky blue': '#75bbfd',
            'slate': '#516572',
            'slate blue': '#5b7c99',
            'slate green': '#658d6d',
            'slate grey': '#59656d',
            'slime green': '#99cc04',
            'snot': '#acbb0d',
            'snot green': '#9dc100',
            'soft blue': '#6488ea',
            'soft green': '#6fc276',
            'soft pink': '#fdb0c0',
            'soft purple': '#a66fb5',
            'spearmint': '#1ef876',
            'spring green': '#a9f971',
            'spruce': '#0a5f38',
            'squash': '#f2ab15',
            'steel': '#738595',
            'steel blue': '#5a7d9a',
            'steel grey': '#6f828a',
            'stone': '#ada587',
            'stormy blue': '#507b9c',
            'straw': '#fcf679',
            'strawberry': '#fb2943',
            'strong blue': '#0c06f7',
            'strong pink': '#ff0789',
            'sun yellow': '#ffdf22',
            'sunflower': '#ffc512',
            'sunflower yellow': '#ffda03',
            'sunny yellow': '#fff917',
            'sunshine yellow': '#fffd37',
            'swamp': '#698339',
            'swamp green': '#748500',
            'tan': '#d1b26f',
            'tan brown': '#ab7e4c',
            'tan green': '#a9be70',
            'tangerine': '#ff9408',
            'taupe': '#b9a281',
            'tea': '#65ab7c',
            'tea green': '#bdf8a3',
            'teal': '#029386',
            'teal blue': '#01889f',
            'teal green': '#25a36f',
            'tealish': '#24bca8',
            'tealish green': '#0cdc73',
            'terra cotta': '#c9643b',
            'terracota': '#cb6843',
            'terracotta': '#ca6641',
            'tiffany blue': '#7bf2da',
            'tomato': '#ef4026',
            'tomato red': '#ec2d01',
            'topaz': '#13bbaf',
            'toupe': '#c7ac7d',
            'toxic green': '#61de2a',
            'tree green': '#2a7e19',
            'true blue': '#010fcc',
            'true green': '#089404',
            'turquoise': '#06c2ac',
            'turquoise blue': '#06b1c4',
            'turquoise green': '#04f489',
            'turtle green': '#75b84f',
            'twilight': '#4e518b',
            'twilight blue': '#0a437a',
            'ugly blue': '#31668a',
            'ugly brown': '#7d7103',
            'ugly green': '#7a9703',
            'ugly pink': '#cd7584',
            'ugly purple': '#a442a0',
            'ugly yellow': '#d0c101',
            'ultramarine': '#2000b1',
            'ultramarine blue': '#1805db',
            'umber': '#b26400',
            'velvet': '#750851',
            'vermillion': '#f4320c',
            'very dark blue': '#000133',
            'very dark brown': '#1d0200',
            'very dark green': '#062e03',
            'very dark purple': '#2a0134',
            'very light blue': '#d5ffff',
            'very light brown': '#d3b683',
            'very light green': '#d1ffbd',
            'very light pink': '#fff4f2',
            'very light purple': '#f6cefc',
            'very pale blue': '#d6fffe',
            'very pale green': '#cffdbc',
            'vibrant blue': '#0339f8',
            'vibrant green': '#0add08',
            'vibrant purple': '#ad03de',
            'violet': '#9a0eea',
            'violet blue': '#510ac9',
            'violet pink': '#fb5ffc',
            'violet red': '#a50055',
            'viridian': '#1e9167',
            'vivid blue': '#152eff',
            'vivid green': '#2fef10',
            'vivid purple': '#9900fa',
            'vomit': '#a2a415',
            'vomit green': '#89a203',
            'vomit yellow': '#c7c10c',
            'warm blue': '#4b57db',
            'warm brown': '#964e02',
            'warm grey': '#978a84',
            'warm pink': '#fb5581',
            'warm purple': '#952e8f',
            'washed out green': '#bcf5a6',
            'water blue': '#0e87cc',
            'watermelon': '#fd4659',
            'weird green': '#3ae57f',
            'wheat': '#fbdd7e',
            'white': '#ffffff',
            'windows blue': '#3778bf',
            'wine': '#80013f',
            'wine red': '#7b0323',
            'wintergreen': '#20f986',
            'wisteria': '#a87dc2',
            'yellow': '#ffff14',
            'yellow brown': '#b79400',
            'yellow green': '#c0fb2d',
            'yellow ochre': '#cb9d06',
            'yellow orange': '#fcb001',
            'yellow tan': '#ffe36e',
            'yellow/green': '#c8fd3d',
            'yellowgreen': '#bbf90f',
            'yellowish': '#faee66',
            'yellowish brown': '#9b7a01',
            'yellowish green': '#b0dd16',
            'yellowish orange': '#ffab0f',
            'yellowish tan': '#fcfc81',
            'yellowy brown': '#ae8b0c',
            'yellowy green': '#bff128'}